import re
import asyncio
from pathlib import Path

from nonebot import logger, require, get_driver, on_command, on_startswith
from nonebot.plugin import PluginMetadata
from nonebot.plugin.load import inherit_supported_adapters

require("nonebot_plugin_uninfo")
require("nonebot_plugin_alconna")
require("nonebot_plugin_apscheduler")
require("nonebot_plugin_localstore")
require("nonebot_plugin_htmlrender")
from nonebot_plugin_apscheduler import scheduler

from .config import CHINESE_FONT_PATH, Config, fconfig

__plugin_meta__ = PluginMetadata(
    name="堡垒之夜游戏插件",
    description="堡垒之夜战绩, 季卡, 商城, vb图查询",
    usage="季卡/生涯季卡/战绩/生涯战绩/商城/vb图",
    type="application",
    config=Config,
    homepage="https://github.com/fllesser/nonebot-plugin-fortnite",
    supported_adapters=inherit_supported_adapters(
        "nonebot_plugin_alconna", "nonebot_plugin_uninfo"
    ),
)

from . import pve, shop, stats, utils


@get_driver().on_startup
async def check_resources():
    import httpx
    import aiofiles

    paths = [CHINESE_FONT_PATH]

    async def dwonload_file(path: Path):
        url = f"{fconfig.raw_base_url}/master/resources/{path.name}"
        logger.info(f"文件 {path.name} 不存在，开始从 {url} 下载...")
        try:
            async with httpx.AsyncClient(timeout=60) as client:
                response = await client.get(url)
            response.raise_for_status()
            font_data = response.content

            async with aiofiles.open(path, "wb") as f:
                await f.write(font_data)

            size = utils.get_size_in_mb(path)
            logger.success(f"文件 {path.name} 下载成功，文件大小: {size:.2f} MB")
        except Exception:
            logger.exception("文件下载失败")
            logger.warning(f"请前往仓库下载资源文件到 {path}")

    await asyncio.gather(*[dwonload_file(path) for path in paths if not path.exists()])


if fconfig.github_token is None:
    minute = 2
else:
    minute = 0


@scheduler.scheduled_job(
    "cron",
    id="fortnite-screenshot",
    hour=8,
    minute=minute,
    coalesce=True,
    misfire_grace_time=60,
)
async def daily_update():
    if fconfig.github_token is not None:
        await utils.trigger_screenshot_action()
        await asyncio.sleep(90)

    logger.info("开始更新商城/VB图...")
    try:
        await shop.update_shop_img()
    except Exception:
        logger.exception("商城更新失败")
    try:
        await pve.update_vb_img()
    except Exception:
        logger.exception("vb图更新失败")


from arclet.alconna import Args, Alconna, Arparma
from nonebot.permission import SUPERUSER
from nonebot_plugin_uninfo import Uninfo
from nonebot_plugin_alconna import Match, AlconnaMatcher, on_alconna
from nonebot_plugin_alconna.uniseg import Text, Image, UniMessage

timewindow_prefix = ["生涯", ""]
name_args = Args["name?", str]

battle_pass_alc = on_alconna(Alconna(timewindow_prefix, "季卡", name_args))
stats_alc = on_alconna(Alconna(timewindow_prefix, "战绩", name_args))


@battle_pass_alc.handle()
@stats_alc.handle()
async def _(matcher: AlconnaMatcher, session: Uninfo, name: Match[str]):
    if name.available:
        matcher.set_path_arg("name", name.result)
        return
    # 获取群昵称
    if not session.member or not session.member.nick:
        return
    pattern = r"(?:id:|id\s)(.+)"
    if matched := re.match(pattern, session.member.nick, re.IGNORECASE):
        matcher.set_path_arg("name", matched.group(1))


name_prompt = UniMessage.template(
    "{:At(user, $event.get_user_id())} 请发送游戏名称\n"
    "群昵称设置如下可快速查询:\n"
    "    id:name\n"
    "    ID name"
)


@battle_pass_alc.got_path("name", prompt=name_prompt)
async def _(arp: Arparma, name: str):
    header: str = arp.header_match.result
    receipt = await UniMessage.text(f"正在查询 {name} 的{header}，请稍后...").send()
    level_info = await stats.get_level(name, header)
    await UniMessage(Text(level_info)).send()
    await receipt.recall(delay=1)


@stats_alc.got_path("name", prompt=name_prompt)
async def _(arp: Arparma, name: str):
    header: str = arp.header_match.result
    receipt = await UniMessage.text(f"正在查询 {name} 的{header}，请稍后...").send()
    try:
        file = await stats.get_stats_image(name, header)
    except Exception as e:
        if isinstance(e, ValueError):
            await UniMessage(Text(str(e))).finish()
        logger.exception("查询失败")
        await UniMessage(Text("查询失败")).finish()
    await UniMessage(Image(path=file)).send()
    await receipt.recall(delay=1)
    file.unlink(missing_ok=True)


shop_matcher = on_startswith("商城")


@shop_matcher.handle()
async def _():
    if not shop.SHOP_FILE.exists():
        logger.info("商城图不存在, 开始更新商城...")
        try:
            await shop.update_shop_img()
            size = utils.get_size_in_mb(shop.SHOP_FILE)
            logger.success(f"商城更新成功，文件大小: {size:.2f} MB")
        except Exception:
            logger.exception("商城更新失败")
    await UniMessage(
        Image(path=shop.SHOP_FILE)
        + Text("可前往 https://www.fortnite.com/item-shop?lang=zh-Hans 购买")
    ).send()


@on_startswith("更新商城", permission=SUPERUSER).handle()
async def _():
    receipt = await UniMessage.text("正在更新商城，请稍后...").send()
    try:
        await shop.update_shop_img()
        await UniMessage(Text("手动更新商城成功") + Image(path=shop.SHOP_FILE)).send()
    except Exception:
        logger.exception("手动更新商城失败")
        await UniMessage(Text("手动更新商城失败")).send()
    finally:
        await receipt.recall(delay=1)


vb_matcher = on_startswith(("vb图", "VB图", "Vb图"))


@vb_matcher.handle()
async def _():
    if not pve.VB_FILE.exists():
        logger.info("vb 图不存在, 开始更新vb图...")
        try:
            await pve.update_vb_img()
            size = utils.get_size_in_mb(pve.VB_FILE)
            logger.success(f"vb图更新成功, 文件大小: {size:.2f} MB")
        except Exception as e:
            logger.warning(f"vb图更新失败: {e}")
    await UniMessage(Image(path=pve.VB_FILE)).send()


@on_startswith("更新vb图", permission=SUPERUSER).handle()
async def _():
    receipt = await UniMessage.text("正在更新vb图, 请稍后...").send()
    try:
        await pve.update_vb_img()
        await UniMessage(Text("手动更新 VB 图成功") + Image(path=pve.VB_FILE)).send()
    except Exception as e:
        await UniMessage(Text(f"手动更新 VB 图失败 | {e}")).send()
    finally:
        await receipt.recall(delay=1)


if fconfig.github_token is not None:
    action_matcher = on_command("更新堡垒", permission=SUPERUSER)

    @action_matcher.handle()
    async def _():
        await utils.trigger_screenshot_action()
        await UniMessage(Text(utils.TRIGGER_SCREENSHOT_TIP)).send()
        await asyncio.sleep(70)

        await shop.update_shop_img()
        await pve.update_vb_img()

        await UniMessage(
            Text("更新成功") + Image(path=shop.SHOP_FILE) + Image(path=pve.VB_FILE)
        ).send()



import asyncio
import hashlib
from io import BytesIO
from typing import Any
from pathlib import Path
from functools import cache

import httpx
import aiofiles
from PIL import Image, ImageDraw, ImageFont
from fortnite_api import Client
from fortnite_api.enums import TimeWindow, StatsImageType
from fortnite_api.errors import FortniteAPIException

from .config import CHINESE_FONT_PATH, fconfig

API_KEY: str | None = fconfig.fortnite_api_key
STATS_BG_PATH = Path(__file__).parent / "resources" / "stats_bg.png"


@cache
def create_gradient_image_new() -> Image.Image:
    """从底图裁剪渐变图片"""
    left, top, right, bottom = 26, 90, 423, 230

    with Image.open(STATS_BG_PATH) as img:
        gradient_img = img.crop((left, top, right, bottom))
        gradient_img.load()
        return gradient_img


def handle_fortnite_api_exception(e: FortniteAPIException) -> str:
    err_msg = str(e)
    if "public" in err_msg:
        return "战绩未公开"
    elif "exist" in err_msg:
        return "用户不存在"
    elif "match" in err_msg:
        return "该玩家当前赛季没有进行过任何对局"
    elif "timed out" in err_msg:
        return "请求超时, 请稍后再试"
    elif "failed to fetch" in err_msg:
        return "拉取账户信息失败, 稍后再试"
    else:
        return f"未知错误: {err_msg}"


async def get_level(name: str, cmd_header: str) -> str:
    time_window: Any = (
        TimeWindow.LIFETIME if cmd_header.startswith("生涯") else TimeWindow.SEASON
    )
    try:
        async with Client(api_key=API_KEY) as client:
            stats = await client.fetch_br_stats(name=name, time_window=time_window)
    except FortniteAPIException as e:
        return handle_fortnite_api_exception(e)
    bp = stats.battle_pass
    if bp is None:
        return f"未查询到 {stats.user.name} 的季卡等级"
    return f"{stats.user.name}: Lv{bp.level} | {bp.progress}% to Lv{bp.level + 1}"


async def get_stats_image(name: str, cmd_header: str) -> Path:
    time_window: Any = (
        TimeWindow.LIFETIME if cmd_header.startswith("生涯") else TimeWindow.SEASON
    )
    image_type: Any = StatsImageType.ALL
    try:
        async with Client(api_key=API_KEY) as client:
            stats = await client.fetch_br_stats(
                name=name,
                time_window=time_window,
                image=image_type,
            )
    except FortniteAPIException as e:
        raise ValueError(handle_fortnite_api_exception(e))
    if stats.image is None:
        raise ValueError(f"未查询到 {stats.user.name} 的战绩")

    img_bytes = await get_stats_img_by_url(stats.image.url, stats.user.name)

    hash_str = hashlib.md5(f"{name}{cmd_header}".encode()).hexdigest()
    file_path = fconfig.data_dir / f"{hash_str}.png"

    async with aiofiles.open(file_path, "wb") as f:
        await f.write(img_bytes.getvalue())

    return file_path


async def get_stats_img_by_url(url: str, name: str) -> BytesIO:
    async with httpx.AsyncClient(verify=False, timeout=15) as client:
        # 发送GET请求获取图片数据
        response = await client.get(url)

        # 检查请求是否成功
        if response.status_code != 200:
            raise ValueError(f"无法获取图片, 状态码: {response.status_code}")

        # 将响应内容转换为字节流
        image_data = BytesIO(response.content)
    # 如果不包含中文名，返回原图
    if not contains_chinese(name):
        return image_data

    return await process_image_with_chinese(image_data, name)


def contains_chinese(text: str) -> bool:
    import re

    pattern = re.compile(r"[\u4e00-\u9fff]")
    return bool(pattern.search(text))


async def process_image_with_chinese(file: BytesIO, name: str) -> BytesIO:
    return await asyncio.to_thread(_process_image_with_chinese, file, name)


def _process_image_with_chinese(bytes_io: BytesIO, name: str) -> BytesIO:
    with Image.open(bytes_io, formats=["PNG"]) as img:
        draw = ImageDraw.Draw(img)

        # 矩形区域的坐标
        left, top, right, bottom = 26, 90, 423, 230

        # 创建渐变色并填充矩形区域
        # width = right - left, height = bottom - top
        gradient_img = create_gradient_image_new()
        img.paste(gradient_img, (left, top))
        # 指定字体
        font_size = 36
        font = ImageFont.truetype(CHINESE_FONT_PATH, font_size)

        # 计算字体坐标
        length = draw.textlength(name, font=font)
        x = left + (right - left - length) / 2
        y = top + (bottom - top - font_size) / 2
        draw.text((x, y), name, fill="#fafafa", font=font)

        output_bytes = BytesIO()
        # 保存处理后的图像到 BytesIO
        img.save(output_bytes, format="PNG", optimize=True)
        # 将指针重置到 BytesIO 对象的开头
        output_bytes.seek(0)

        return output_bytes



from pathlib import Path

import nonebot_plugin_localstore as store
from nonebot import get_plugin_config
from pydantic import BaseModel

_cache_dir: Path = store.get_plugin_cache_dir()
_data_dir: Path = store.get_plugin_data_dir()

CHINESE_FONT_PATH: Path = _data_dir / "SourceHanSansSC-Bold-2.otf"
GITHUB_RAW = "https://raw.githubusercontent.com/fllesser/nonebot-plugin-fortnite"


class Config(BaseModel):
    fortnite_api_key: str | None = None
    fortnite_screenshot_from_github: bool = False
    fortnite_github_proxy_url: str | None = None  # "https://gh-proxy.org"
    fortnite_github_token: str | None = None

    @property
    def api_key(self) -> str | None:
        return self.fortnite_api_key

    @property
    def screenshot_from_github(self) -> bool:
        return self.fortnite_screenshot_from_github

    @property
    def raw_base_url(self) -> str:
        if self.fortnite_github_proxy_url is None:
            return GITHUB_RAW

        proxy = self.fortnite_github_proxy_url.rstrip("/")
        return f"{proxy}/{GITHUB_RAW}"

    @property
    def cache_dir(self) -> Path:
        return _cache_dir

    @property
    def data_dir(self) -> Path:
        return _data_dir

    @property
    def chinese_font_path(self) -> Path:
        return CHINESE_FONT_PATH

    @property
    def github_token(self) -> str | None:
        return self.fortnite_github_token


fconfig: Config = get_plugin_config(Config)



import time
import asyncio
from pathlib import Path
from contextlib import ExitStack

import httpx
import aiofiles
from PIL import Image, ImageDraw, ImageFont
from nonebot.log import logger
from playwright.async_api import Route
from nonebot_plugin_htmlrender import get_new_page
from nonebot_plugin_htmlrender.browser import Page

from . import utils
from .config import fconfig

VB_FILE_NAME = "vb.png"
VB_FILE = fconfig.data_dir / VB_FILE_NAME
VB_FONT_PATH = Path(__file__).parent / "resources" / "LuckiestGuy.woff"


async def update_vb_img():
    """更新 VB 图片（根据配置决定下载或截图）"""

    if fconfig.fortnite_screenshot_from_github:
        logger.info("从 GitHub 分支下载 VB 图片...")
        await download_vb_img_from_github()
    else:
        await screenshot_vb_img()

    size = utils.get_size_in_mb(VB_FILE)
    logger.success(f"vb图更新成功, 文件大小: {size:.2f} MB")


@utils.retry(retries=3, delay=10)
async def download_vb_img_from_github():
    """从 GitHub 分支下载 VB 图片"""

    url = f"{fconfig.raw_base_url}/screenshots/{VB_FILE_NAME}"

    async with httpx.AsyncClient(timeout=30) as client:
        response = await client.get(url)
        response.raise_for_status()

    async with aiofiles.open(VB_FILE, "wb") as f:
        await f.write(response.content)


async def screenshot_vb_img():
    async with get_new_page(device_scale_factor=1) as page:
        await _screenshot_vb_img(page)
    await combine_imgs()


_SELECTOR_MAP = {
    "hot_info_1.png": ("div.hot-info", 0),
    "container_hidden_xs.png": ("div.container.hidden-xs", 0),
    "hot_info_2.png": ("div.hot-info", 1),
}


@utils.retry(retries=3, delay=10)
async def _screenshot_vb_img(page: Page):
    url = "https://freethevbucks.com/timed-missions"

    # 拦截广告
    async def ad_block_handler(route: Route):
        ad_domains = [
            "googlesyndication.com",
            "doubleclick.net",
            "adnxs.com",
            "google-analytics.com",
            "facebook.com",
            "amazon-adsystem.com",
            "adform.net",
            "googleadservices.com",
            "doubleclick.net",
        ]
        if any(ad_domain in route.request.url for ad_domain in ad_domains):
            await route.abort()
        else:
            await route.continue_()

    await page.route("**/*", ad_block_handler)
    await page.goto(url)

    # 截图
    async def screenshot(filename: str, selector: str, nth: int = 0) -> None:
        locator = page.locator(selector).nth(nth)
        # 检查元素内容是否为空
        content = await locator.inner_html()
        path = fconfig.cache_dir / filename
        if content.strip():
            await asyncio.wait_for(locator.screenshot(path=path), timeout=5)
        else:
            logger.warning(f"Locator for {path.name} is empty.")

    await asyncio.gather(*[screenshot(file, *sn) for file, sn in _SELECTOR_MAP.items()])


async def combine_imgs():
    await asyncio.to_thread(_combine_imgs)


def _combine_imgs():
    # 打开截图文件（如果存在）
    img_paths = [fconfig.cache_dir / file for file in _SELECTOR_MAP.keys()]
    img_paths = [path for path in img_paths if path.exists()]
    if not img_paths:
        raise Exception("所有选择器的截图文件均不存在")
    # 先添加时间
    try:
        with ExitStack() as stack:
            # 动态打开所有图片
            images: list[Image.Image] = [
                stack.enter_context(Image.open(path)) for path in img_paths
            ]

            # 获取尺寸并创建新图像
            widths, heights = zip(*(img.size for img in images))
            total_width = max(widths)
            total_height = sum(heights)

            # 如果 img1.width < total_width，则拉伸最右侧像素到 total_width
            if (img1 := images[0]) and img1.width < total_width:
                images[0] = resize_img_with_right_pixel(img1, total_width)

            # 填充更新时间
            draw_time_text(img1, total_width)
            with Image.new("RGB", (total_width, total_height)) as combined_image:
                # 将截图粘贴到新图像中
                y_offset = 0
                for img in images:
                    combined_image.paste(img, (0, y_offset))
                    y_offset += img.height

                # 保存合并后的图像
                combined_image.save(VB_FILE)
    finally:
        # 关闭并删除所有截图文件
        for img_path in img_paths:
            img_path.unlink()


def draw_time_text(img: Image.Image, width: int = 1126):
    draw = ImageDraw.Draw(img)
    font_size = 26
    font = ImageFont.truetype(VB_FONT_PATH, font_size)
    time_text = time.strftime("Updated: %Y-%m-%d %H:%M:%S", time.localtime())
    time_text_width = draw.textlength(time_text, font=font)
    x = width - time_text_width - 10
    draw.text((x, 12), time_text, font=font, fill=(80, 80, 80))


def resize_img_with_right_pixel(img: Image.Image, width: int = 1126):
    new_img = Image.new("RGB", (width, img.height))
    new_img.paste(img, (0, 0))
    # 横向取 img 最右侧像素点，填充到 new_img 的 width - 1 到 width 的像素点
    for x in range(img.width - 50, width):
        for y in range(img.height):
            color = img.getpixel((img.width - 50, y))
            assert color is not None
            new_img.putpixel((x, y), color)
    return new_img



import asyncio
from pathlib import Path

import httpx
import aiofiles
from PIL import Image
from nonebot import logger

from .config import fconfig

TRIGGER_SCREENSHOT_TIP = (
    "Trigger screenshot action, "
    "https://github.com/fllesser/nonebot-plugin-fortnite/actions/workflows/screenshot.yml"
)


async def trigger_screenshot_action():
    url = "https://api.github.com/repos/fllesser/nonebot-plugin-fortnite/actions/workflows/screenshot.yml/dispatches"
    headers = {"Authorization": f"token {fconfig.github_token}"}
    async with httpx.AsyncClient(headers=headers) as client:
        response = await client.post(url, json={"ref": "master"})
        response.raise_for_status()

    logger.info(TRIGGER_SCREENSHOT_TIP)


async def save_img(img: Image.Image, path: Path, format: str = "PNG"):
    from io import BytesIO

    buffer = BytesIO()
    img.save(buffer, format=format)
    img_data = buffer.getvalue()
    async with aiofiles.open(path, "wb") as f:
        await f.write(img_data)


def get_size_in_mb(path: Path):
    return path.stat().st_size / 1024 / 1024


def retry(retries: int = 3, delay: float = 3):
    from functools import wraps

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            for i in range(retries):
                try:
                    return await func(*args, **kwargs)
                except Exception:
                    logger.warning(f"Error in {func.__name__}, retry {i + 1}/{retries}")
                    if i == retries - 1:
                        raise
                    await asyncio.sleep(delay)

        return wrapper

    return decorator



import asyncio
from pathlib import Path

from nonebot import logger
from nonebot_plugin_htmlrender import get_new_page
from nonebot_plugin_htmlrender.browser import Page

from . import utils
from .config import fconfig

SHOP_FILE_NAME = "shop.png"
SHOP_FILE = fconfig.data_dir / SHOP_FILE_NAME
GG_FONT_PATH = Path(__file__).parent / "resources" / "burbankbigregular-black.woff2"


async def update_shop_img():
    """更新商城图片（根据配置决定下载或截图）"""

    if fconfig.fortnite_screenshot_from_github:
        logger.info("从 GitHub 下载商城图片...")
        await download_shop_img_from_github()
    else:
        await screenshot_shop_img()

    size = utils.get_size_in_mb(SHOP_FILE)
    logger.success(f"商城更新成功，文件大小: {size:.2f} MB")


@utils.retry(retries=3, delay=10)
async def download_shop_img_from_github():
    """从 GitHub 分支下载商城图片"""
    import httpx
    import aiofiles

    url = f"{fconfig.raw_base_url}/screenshots/{SHOP_FILE_NAME}"

    async with httpx.AsyncClient(timeout=60) as client:
        async with client.stream("GET", url) as response:
            response.raise_for_status()
            # 流式写入文件
            async with aiofiles.open(SHOP_FILE, "wb") as f:
                async for chunk in response.aiter_bytes(8192):
                    await f.write(chunk)


async def screenshot_shop_img():
    # url = "https://www.fortnite.com/item-shop?lang=zh-Hans"
    headers = {
        "User-Agent": (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
            " (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36"
        ),
        "Accept": (
            "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,"
            "image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9"
        ),
        "Accept-Encoding": "gzip, deflate",
        "upgrade-insecure-requests": "1",
        "dnt": "1",
        "x-requested-with": "mark.via",
        "sec-fetch-site": "none",
        "sec-fetch-mode": "navigate",
        "sec-fetch-user": "?1",
        "sec-fetch-dest": "document",
        "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
    }
    # browser = await get_browser(headless=True)
    # context = await browser.new_context(extra_http_headers=headers)
    async with get_new_page(device_scale_factor=1, extra_http_headers=headers) as page:
        await _screenshot_shop_img(page)
    await add_update_time()


@utils.retry(3, 10)
async def _screenshot_shop_img(page: Page):
    url = "https://fortnite.gg/shop"
    await page.add_style_tag(
        content="* { transition: none !important; animation: none !important; }"
    )
    await page.goto(url)

    async def wait_for_load():
        await page.wait_for_load_state("networkidle", timeout=90000)

    async def scroll_page():
        for _ in range(20):
            await page.evaluate("""() => {
                window.scrollBy(0, document.body.scrollHeight / 20);
            }""")
            await asyncio.sleep(0.5)  # 等待 0.5 秒以加载内容

    await asyncio.gather(wait_for_load(), scroll_page())
    await page.screenshot(path=SHOP_FILE, full_page=True)


async def add_update_time():
    await asyncio.to_thread(_add_update_time)


def _add_update_time():
    import time

    from PIL import Image, ImageDraw, ImageFont

    font = ImageFont.truetype(GG_FONT_PATH, 88)
    with Image.open(SHOP_FILE) as img:
        draw = ImageDraw.Draw(img)
        # 先填充 rgb(47,49,54) 背景 1280 * 100
        draw.rectangle((0, 0, 1280, 270), fill=(47, 49, 54))
        # 1280 宽，19个数字居中 x 坐标
        time_text = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        time_text_width = draw.textlength(time_text, font=font)
        x = (1280 - time_text_width) / 2
        draw.text((x, 100), time_text, font=font, fill=(255, 255, 255))
        img.save(SHOP_FILE)
